from hier_config import HConfig
import hier_config.helpers as H


class Host:
    """
    A host object is a convenient way to loading host inventory
    items into a single object.

    The default is to load "hostname", "os", and "options" to the host object,
    however, it can easily be extended for developer needs.

    .. code:: python

        import yaml
        from hier_config.host import Host

        options = yaml.load(open('./tests/files/test_options_ios.yml'))
        host = Host('example.rtr', 'ios', options)

        # Example of easily extending the host object
        host.facts['chassis_model'] = 'WS-C4948E'

        # Example of loading running config and compiled configs into a host object
        host.load_config_from(config_type="running", name="./tests/files/running_config.conf")
        host.load_config_from(config_type="compiled", name="./tests/files/compiled_config.conf")

        # Example of loading hier-config tags into a host object
        host.load_tags("./tests/files/test_tags_ios.yml")

        # Example of creating a remediation config without a tag targeting specific config
        host.load_remediation()

        # Example of creating a remediation config with a tag ('safe') targeting a specific config.
        host.filter_remediation(include_tags=['safe'])

    :param hostname: type str
    :param os: type str
    :param hconfig_options: type dict

    :return: Host Object

    """

    def __init__(self, hostname, os, hconfig_options):
        self.hostname = str(hostname)
        self.os = str(os)
        self.hconfig_options = dict(hconfig_options)
        self._hconfig_tags = list()
        self._running_config = None
        self._compiled_config = None
        self._remediation_config = None
        self.facts = dict()

    def __repr__(self):
        return 'Host(hostname={})'.format(self.hostname)

    @property
    def running_config(self):
        """
        running configuration property

        :return: self._running_config -> type HConfig Object or None
        """
        if self._running_config is None:
            self._running_config = self._get_running_config()
        return self._running_config

    @property
    def compiled_config(self):
        """
        compiled configuration property

        :return: self._compiled_config -> type HConfig Object or None
        """
        if self._compiled_config is None:
            self._compiled_config = self._get_compiled_config()
        return self._compiled_config

    @property
    def remediation_config(self):
        """
        remediation configuration property

        :return: self._remediation_config -> type HConfig Object or None
        """
        if self._remediation_config is None:
            self._remediation_config = self._get_remediation_config()
        return self._remediation_config

    @property
    def hconfig_tags(self):
        """
        hier-config tags property

        :return: self._hconfig_tags -> type list of dicts
        """
        return self._hconfig_tags

    def load_config_from(self, config_type, name, load_file=True):
        """
        1. Loads a running config or a compiled config into a Host object
        2. Sets host.facts['running_config_raw'] or host.facts['compiled_config_raw']
        3. Loads the config into HConfig
        4. Sets the loaded hier-config in host.facts['running_config'] or host.facts['compiled_config']

        :param config_type: 'running' or 'compiled' -> type str
        :param name: file name or config text string to load -> type str
        :param load_file: default, True -> type bool
        :return: self.running_config or self.compiled_config
        """
        hier = HConfig(host=self)

        if load_file:
            config_text = self._load_from_file(name)
        else:
            config_text = name

        hier.load_from_string(config_text)

        if config_type == "running":
            self.facts["running_config_raw"] = config_text
            self._running_config = hier

            return self.running_config
        elif config_type == "compiled":
            self.facts["compiled_config_raw"] = config_text
            self._compiled_config = hier

            return self.compiled_config
        else:
            raise SyntaxError("Unknown config_type. Expected 'running' or 'compiled'")

    def load_remediation(self):
        """
        Once self.running_config and self.compled_config have been created,
        create self.remediation_config

        :return: self.remediation_config
        """
        if self.running_config and self.compiled_config:
            self._remediation_config = self.running_config.config_to_get_to(
                self.compiled_config
            )
        else:
            raise AttributeError("Missing host.running_config or host.compiled_config")

        self.remediation_config.add_sectional_exiting()
        self.remediation_config.set_order_weight()
        self.remediation_config.add_tags(self.hconfig_tags)
        self.filter_remediation()

        return self.remediation_config

    def filter_remediation(self, include_tags=None, exclude_tags=None):
        """
        Run filter jobs, based on tags on self.remediation_config

        :param include_tags: type list
        :param exclude_tags: type list
        :return: self.facts['remediation_conig_raw'] -> type str
        """
        remediation_text = str()

        if include_tags or exclude_tags is not None:
            include_tags = H.to_list(include_tags)
            exclude_tags = H.to_list(exclude_tags)

            for line in self.remediation_config.all_children_sorted_by_tags(include_tags, exclude_tags):
                remediation_text += line.cisco_style_text()
                remediation_text += '\n'
        else:
            for line in self.remediation_config.all_children():
                remediation_text += line.cisco_style_text()
                remediation_text += '\n'

        self.facts["remediation_config_raw"] = remediation_text

        return self.facts["remediation_config_raw"]

    def load_tags(self, name, load_file=True):
        """
        Loads lineage rules into host.facts["hconfig_tags"]

        Example:
            Specify to load lineage rules from a file.

        .. code:: python

            host.load_tags('tags_ios.yml')

        Example:
            Specify to load lineage rules from a dictionary.

        .. code:: python

            tags = [{"lineage": [{"startswith": "interface"}], "add_tags": "interfaces"}]
            host.load_tags(tags, file=False)

        :param name: tags from a file or dictionary
        :param load_file: default, True -> type bool
        :return: self.hconfig_tags
        """
        if load_file:
            self._hconfig_tags = self._load_from_file(name, parse_yaml=True)
        else:
            self._hconfig_tags = name

        return self.hconfig_tags

    @staticmethod
    def _load_from_file(name, parse_yaml=False):
        """
        Opens a config file and loads it as a string.

        :param name: type str
        :param parse_yaml: type boolean
        :return: content -> type str or type dict
        """
        with open(name) as f:
            content = f.read()

        if parse_yaml:
            import yaml
            content = yaml.safe_load(content)

        return content

    def _get_running_config(self):
        return NotImplemented

    def _get_compiled_config(self):
        return NotImplemented

    def _get_remediation_config(self):
        return NotImplemented



from hier_config.text_match import TextMatch

import hier_config.helpers as H


class HConfigChild:

    def __init__(self, parent, text):
        self.parent = parent
        self._text = text.strip()
        self.real_indent_level = None
        self.children = []
        self.children_dict = {}
        # The intent is for self.order_weight values to range from 1 to 999
        # with the default weight being 500
        self.order_weight = 500
        self.tags = set()
        self.comments = set()
        self.new_in_config = False
        self.instances = []

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, value):
        """
        Used for when self.text is changed after the object
        is instantiated to rebuild the children dictionary

        """

        self._text = value.strip()
        self.parent.rebuild_children_dict()

    def __repr__(self):
        if self.parent is self.root:
            return 'HConfigChild(HConfig, {})'.format(self.text)
        return 'HConfigChild(HConfigChild, {})'.format(self.text)

    def __str__(self):
        return self.text

    def __lt__(self, other):
        return self.order_weight < other.order_weight

    def __len__(self):
        length = 0
        for _ in self.all_children():
            length += 1
        return length

    def __bool__(self):
        return True

    def __contains__(self, item):
        return str(item) in self.children_dict

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        attributes_to_compare = (
            'text',
            'tags',
            'comments',
            'new_in_config',
        )

        for attribute in attributes_to_compare:
            self_attribute = getattr(self, attribute)
            other_attribute = getattr(other, attribute)
            if self_attribute != other_attribute:
                return False

        if len(self.children) != len(other.children):
            return False

        for self_child, other_child in zip(sorted(self.children), sorted(other.children)):
            if self_child != other_child:
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def root(self):
        """ returns the HConfig object at the base of the tree """

        return self.parent.root

    @property
    def logs(self):
        return self.root.logs

    @property
    def host(self):
        return self.root.host

    @property
    def options(self):
        return self.root.options

    @staticmethod
    def _lineage_eval_object_rules(rules, section):
        """
        Evaluate a list of lineage object rules.

        All object rules must match in order to return True

        """

        matches = 0
        for rule in rules:
            if rule['test'] == 'new_in_config':
                if rule['expression'] == section.new_in_config:
                    matches += 1
            elif rule['test'] == 'negative_intersection_tags':
                rule['expression'] = H.to_list(rule['expression'])
                if not set(rule['expression']).intersection(section.tags):
                    matches += 1
        return matches == len(rules)

    @staticmethod
    def _lineage_eval_text_match_rules(rules, text):
        """
        Evaluate a list of lineage text_match rules.

        Only one text_match rule must match in order to return True

        """

        for rule in rules:
            if TextMatch.dict_call(rule['test'], text, rule['expression']):
                return True
        return False

    @staticmethod
    def _explode_lineage_rule(rule):
        text_match_rules = list()
        object_rules = list()
        for k, v in rule.items():
            if k in ['new_in_config', 'negative_intersection_tags']:
                object_rules.append({'test': k, 'expression': v})
            elif isinstance(v, list):
                text_match_rules += [{'test': k, 'expression': e} for e in v]
            else:
                text_match_rules += [{'test': k, 'expression': v}]
        return object_rules, text_match_rules

    def has_children(self):
        return bool(self.children)

    def depth(self):
        return self.parent.depth() + 1

    def get_child(self, test, expression):
        """ Find a child by TextMatch rule. If it is not found, return None """

        if test == 'equals':
            return self.children_dict.get(expression, None)
        else:
            try:
                return next(self.get_children(test, expression))
            except StopIteration:
                return None

    def get_child_deep(self, test_expression_pairs):
        """
        Find a child recursively with a list of test/expression pairs

        e.g.

        .. code:: python

            result = hier_obj.get_child_deep([('equals', 'control-plane'),
                                              ('equals', 'service-policy input system-cpp-policy')])

        Returns:

            HConfigChild or None

        """

        test, expression = test_expression_pairs.pop(0)
        if test == 'equals':
            result = self.children_dict.get(expression, None)
            if result and test_expression_pairs:
                return result.get_child_deep(test_expression_pairs)
            return result
        else:
            try:
                result = next(self.get_children(test, expression))
                if result and test_expression_pairs:
                    return result.get_child_deep(test_expression_pairs)
                return result
            except StopIteration:
                return None

    def get_children(self, test, expression):
        """ Find all children matching a TextMatch rule and return them. """

        for child in self.children:
            if TextMatch.dict_call(test, child.text, expression):
                yield child

    def move(self, new_parent):
        """
        move one HConfigChild object to different HConfig parent object

        .. code:: python

            hier1 = HConfig(host=host)
            interface1 = hier1.add_child('interface Vlan2')
            interface1.add_child('ip address 10.0.0.1 255.255.255.252')

            hier2 = Hconfig(host)

            interface1.move(hier2)

        :param new_parent: HConfigChild object -> type list
        :return: None

        """

        new_parent.children.append(self)
        new_parent.rebuild_children_dict()
        self.delete()

    def del_child_by_text(self, text):
        """ Delete all children with the provided text """

        if text in self.children_dict:
            self.children[:] = [c for c in self.children if c.text != text]
            self.rebuild_children_dict()

    def del_child(self, child):
        """
        Delete a child from self.children and self.children_dict

        .. code:: python
            hier = HConfig(host=host)
            hier.add_child('interface Vlan2')

            hier.del_child(hier.get_child('startswith', 'interface'))

        :param child: HConfigChild object
        :return: None

        """

        try:
            self.children.remove(child)
        except ValueError:
            pass
        else:
            self.rebuild_children_dict()

    def rebuild_children_dict(self):
        """ Rebuild self.children_dict """

        self.children_dict = {}
        for child in self.children:
            self.children_dict[child.text] = self.children_dict.get(
                child.text, child)

    def add_children(self, lines):
        """
        Add child instances of HConfigChild

        :param lines: HConfigChild object -> type list
        :return: None

        """

        if isinstance(lines, str):
            lines = [lines]

        for line in lines:
            self.add_child(line)

    def add_child(self, text, alert_on_duplicate=False, idx=None, force_duplicate=False):
        """ Add a child instance of HConfigChild """

        if idx is None:
            idx = len(self.children)
        # if child does not exist
        if text not in self:
            new_item = HConfigChild(self, text)
            self.children.insert(idx, new_item)
            self.children_dict[text] = new_item
            return new_item
        # if child does exist and is allowed to be installed as a duplicate
        elif self._duplicate_child_allowed_check() or force_duplicate:
            new_item = HConfigChild(self, text)
            self.children.insert(idx, new_item)
            self.rebuild_children_dict()
            return new_item
        else:
            # If the child is already present and the parent does not allow
            # duplicate children, return the existing child
            # Ignore duplicate remarks in ACLs
            if alert_on_duplicate and not text.startswith('remark '):
                if self is self.root:
                    path = [text]
                else:
                    path = list(self.path()) + [text]
                self.logs.append("Found a duplicate section: {}".format(path))
            return self.get_child('equals', text)

    def add_deep_copy_of(self, child_to_add, merged=False):
        """
        Add a nested copy of a child to self

        :param child_to_add: type HConfigCHild
        :param merged: type boolean, default False

        :return: new_child

        """

        new_child = self.add_shallow_copy_of(child_to_add, merged=merged)
        for child in child_to_add.children:
            new_child.add_deep_copy_of(child, merged=merged)

        return new_child

    def lineage(self):
        """
        Return the lineage of parent objects, up to but excluding the root

        """

        if self.root is self.parent:
            yield self
        else:
            for parent in self.parent.lineage():
                yield parent
            yield self

    def path(self):
        """
        Return a list of the text instance variables from self.lineage

        """

        for obj in self.lineage():
            yield obj.text

    def cisco_style_text(self, style='without_comments', tag=None):
        """ Return a Cisco style formated line i.e. indentation_level + text ! comments """

        comments = []
        if style == 'without_comments':
            pass
        elif style == 'merged':
            # count the number of instances that have the tag
            instance_count = 0
            instance_comments = set()
            for instance in self.instances:
                if tag is None or tag in instance['tags']:
                    instance_count += 1
                    instance_comments.update(instance['comments'])

            # should the word 'instance' be plural?
            word = 'instance' if instance_count == 1 else 'instances'

            comments.append('{} {}'.format(
                instance_count,
                word))
            comments.extend(instance_comments)
        elif style == 'with_comments':
            comments.extend(self.comments)

        return "{}{}{}".format(
            "  " * (self.depth() - 1),  # render the indentation
            self.text,
            ' !{}'.format(', '.join(sorted(comments))) if comments else '')

    def all_children_sorted_untagged(self):
        """ Yield all children recursively that are untagged """

        for child in self.all_children_sorted():
            if not child.tags:
                yield child

    def all_children_sorted_by_tags(self, include_tags, exclude_tags):
        """ Yield all children recursively that match include/exlcude tags """

        for child in self.all_children_sorted():
            if child.line_inclusion_test(include_tags, exclude_tags):
                yield child

    def all_children_sorted(self):
        """ Recursively find and yield all children sorted at each hierarchy """

        for child in sorted(self.children):
            yield child
            for sub_child in child.all_children_sorted():
                yield sub_child

    def all_children(self):
        """ Recursively find and yield all children """

        for child in self.children:
            yield child
            for sub_child in child.all_children():
                yield sub_child

    def delete(self):
        """ Delete the current object from its parent """

        # TODO find a way to remove this when sub-classing in HCRoot
        self.parent.del_child(self)

    def set_order_weight(self):
        """
        Sets self.order integer on all children

        """

        for child in self.all_children():
            for rule in self.options['ordering']:
                if child.lineage_test(rule):
                    child.order_weight = rule['order']

    def add_sectional_exiting(self):
        """
        Adds the sectional exiting text as a child

        """

        # TODO why do we need to delete the delete the sub_child and then
        # recreate it?
        for child in self.all_children():
            for rule in self.options['sectional_exiting']:
                if child.lineage_test(rule):
                    if rule['exit_text'] in child:
                        child.del_child_by_text(rule['exit_text'])

                    new_child = child.add_child(rule['exit_text'])
                    new_child.order_weight = 999

    def to_tag_spec(self, tags):
        """
        Returns the configuration as a tag spec definition

        This is handy when you have a segment of config and need to
        generate a tag spec to tag configuration in another instance

        """

        tag_spec = []
        for child in self.all_children():
            if not child.children:
                child_spec = [{'equals': t} for t in child.path()]
                tag_spec.append({'section': child_spec, 'add_tags': tags})
        return tag_spec

    def ancestor_append_tags(self, tags):
        """
        Append tags to self.tags and for all ancestors

        """

        for ancestor in self.lineage():
            ancestor.append_tags(tags)

    def ancestor_remove_tags(self, tags):
        """
        Remove tags to self.tags and for all ancestors

        """

        for ancestor in self.lineage():
            ancestor.remove_tags(tags)

    def deep_append_tags(self, tags):
        """
        Append tags to self.tags and recursively for all children

        """

        self.append_tags(tags)
        for child in self.all_children():
            child.append_tags(tags)

    def deep_remove_tags(self, tags):
        """
        Remove tags from self.tags and recursively for all children

        """

        self.remove_tags(tags)
        for child in self.all_children():
            child.remove_tags(tags)

    def append_tags(self, tags):
        """
        Add tags to self.tags

        """

        tags = H.to_list(tags)
        # self._tags.update(tags)
        self.tags.update(tags)

    def remove_tags(self, tags):
        """
        Remove tags from self.tags

        """

        tags = H.to_list(tags)
        # self._tags.difference_update(tags)
        self.tags.difference_update(tags)

    def with_tags(self, tags, new_instance=None):
        """
        Returns a new instance containing only sub-objects
        with one of the tags in tags

        """

        from hier_config import HConfig
        if new_instance is None:
            new_instance = HConfig(host=self.host)

        for child in self.children:
            if tags.intersection(self.tags):
                new_child = new_instance.add_shallow_copy_of(child)
                child.with_tags(tags, new_instance=new_child)

        return new_instance

    def negate(self):
        """ Negate self.text """

        for rule in self.options['negation_negate_with']:
            if self.lineage_test(rule):
                self.text = rule['use']
                return self

        for rule in self.options['negation_default_when']:
            if self.lineage_test(rule):
                return self._default()

        return self._swap_negation()

    def config_to_get_to(self, target, delta=None):
        """
        Figures out what commands need to be executed to transition from self to target.
        self is the source data structure(i.e. the running_config),
        target is the destination(i.e. compiled_config)

        """

        from hier_config import HConfig
        if delta is None:
            delta = HConfig(host=self.host)

        self._config_to_get_to_left(target, delta)
        self._config_to_get_to_right(target, delta)

        return delta

    def _config_to_get_to_left(self, target, delta):
        # find self.children that are not in target.children - i.e. what needs to be negated or defaulted
        # Also, find out if another command in self.children will overwrite -
        # i.e. be idempotent
        for self_child in self.children:
            if self_child in target:
                continue
            elif self_child.is_idempotent_command(target.children):
                continue
            else:
                # in other but not self
                # add this node but not any children
                deleted = delta.add_child(self_child.text)
                deleted.negate()
                if self_child.children:
                    deleted.comments.add(
                        "removes {} lines".format(len(self_child.children_dict) + 1))

    def _config_to_get_to_right(self, target, delta):
        # find what would need to be added to source_config to get to self
        for target_child in target.children:
            # if the child exist, recurse into its children
            self_child = self.get_child('equals', target_child.text)
            if self_child:
                # This creates a new HConfigChild object just in case there are some delta children
                # Not very efficient, think of a way to not do this
                subtree = delta.add_child(target_child.text)
                self_child.config_to_get_to(target_child, subtree)
                if not subtree.children:
                    subtree.delete()
                # Do we need to rewrite the child and its children as well?
                elif self_child.sectional_overwrite_check():
                    target_child.overwrite_with(self_child, delta, True)
                elif self_child.sectional_overwrite_no_negate_check():
                    target_child.overwrite_with(self_child, delta, False)
            # else the child is absent, add it
            else:
                new_item = delta.add_deep_copy_of(target_child)
                # mark the new item and all of its children as new_in_config
                new_item.new_in_config = True
                for child in new_item.all_children():
                    child.new_in_config = True
                if new_item.children:
                    new_item.comments.add("new section")

    def _swap_negation(self):
        """ Swap negation of a self.text """

        if self.text.startswith('no '):
            self.text = self.text[3:]
        else:
            self.text = 'no ' + self.text
        return self

    def _default(self):
        """ Default self.text """

        if self.text.startswith('no '):
            self.text = 'default ' + self.text[3:]
        else:
            self.text = 'default ' + self.text
        return self

    def _idempotent_acl_check(self):
        """
        Handle conditional testing to determine if idempotent acl handling for iosxr should be used

        :return: boolean

        """

        if self.host.os in {'iosxr'}:
            if self.parent is not self.root:
                acl = ('ipv4 access-list ', 'ipv6 access-list ')
                if self.parent.text.startswith(acl):
                    return True
        return False

    def is_idempotent_command(self, other_children):
        """
        Determine if self.text is an idempotent change.

        :param other_children: HConfigChild object -> type list
        :return: boolean

        """

        # Blacklist commands from matching as idempotent
        for rule in self.options[
                'idempotent_commands_blacklist']:
            if self.lineage_test(rule, True):
                return False

        # Handles idempotent acl entry identification
        if self._idempotent_acl_check():
            if self.host.os in {'iosxr'}:
                self_sn = self.text.split(' ', 1)[0]
                for other_child in other_children:
                    other_sn = other_child.text.split(' ', 1)[0]
                    if self_sn == other_sn:
                        return True

        # Idempotent command identification
        for rule in self.options['idempotent_commands']:
            if self.lineage_test(rule, True):
                for other_child in other_children:
                    if other_child.lineage_test(rule, True):
                        return True

        return False

    def sectional_overwrite_no_negate_check(self):
        """
        Check self's text to see if negation should be handled by
        overwriting the section without first negating it

        """

        for rule in self.options[
                'sectional_overwrite_no_negate']:
            if self.lineage_test(rule):
                return True
        return False

    def sectional_overwrite_check(self):
        """ Determines if self.text matches a sectional overwrite rule """

        for rule in self.options['sectional_overwrite']:
            if self.lineage_test(rule):
                return True
        return False

    def overwrite_with(self, other, delta, negate=True):
        """ Deletes delta.child[self.text], adds a deep copy of self to delta """

        if other.children != self.children:
            if negate:
                delta.del_child_by_text(self.text)
                deleted = delta.add_child(self.text).negate()
                deleted.comments.add("dropping section")
            if self.children:
                delta.del_child_by_text(self.text)
                new_item = delta.add_deep_copy_of(self)
                new_item.comments.add("re-create section")
        return delta

    def _duplicate_child_allowed_check(self):
        """ Determine if duplicate(identical text) children are allowed under the parent """

        for rule in self.options[
                'parent_allows_duplicate_child']:
            if self.lineage_test(rule):
                return True
        return False

    def add_shallow_copy_of(self, child_to_add, merged=False):
        """ Add a nested copy of a child_to_add to self.children """

        new_child = self.add_child(child_to_add.text)

        if merged:
            new_child.instances.append({
                'hostname': child_to_add.host.hostname,
                'comments': child_to_add.comments,
                'tags': child_to_add.tags})
        new_child.comments.update(child_to_add.comments)
        new_child.tags.update(child_to_add.tags)
        new_child.order_weight = child_to_add.order_weight

        return new_child

    def line_inclusion_test(self, include_tags, exclude_tags):
        """
        Given the line_tags, include_tags, and exclude_tags,
        determine if the line should be included

        """

        include_line = False
        if include_tags:
            set_include_tags = set(include_tags)
            include_line = bool(self.tags.intersection(set_include_tags))

        if include_line:
            set_exclude_tags = set(exclude_tags)
            return not bool(self.tags.intersection(set_exclude_tags))

    def lineage_test(self, rule, strip_negation=False):
        """ A generic test against a lineage of HConfigChild objects """

        if 'match_leaf' in rule and rule['match_leaf']:
            # stupid trick to make a generator with one item
            lineage_obj = (self for _ in [None])
            lineage_depth = 1
        else:
            lineage_obj = self.lineage()
            lineage_depth = self.depth()

        if not len(rule['lineage']) == lineage_depth:
            return False

        matches = 0

        for lineage_rule, section in zip(rule['lineage'], lineage_obj):
            object_rules, text_match_rules = HConfigChild._explode_lineage_rule(
                lineage_rule)

            if not HConfigChild._lineage_eval_object_rules(
                    object_rules, section):
                return False

            # This removes negations for each section but honestly,
            # we really only need to do this on the last one
            if strip_negation:
                if section.text.startswith('no '):
                    text = section.text[3:]
                elif section.text.startswith('default '):
                    text = section.text[8:]
                else:
                    text = section.text
            else:
                text = section.text

            if HConfigChild._lineage_eval_text_match_rules(
                    text_match_rules, text):
                matches += 1
                continue
            else:
                return False

        return matches == len(rule['lineage'])



from hier_config.hc_child import HConfigChild

import re

__version__ = '1.4.1'


class HConfig(HConfigChild):

    """
    A class for representing and comparing Cisco configurations in a
    hierarchical tree data structure.

    Example usage:

    .. code:: python

        # Setup basic environment

        from hier_config import HConfig
        from hier_config.host import Host
        import yaml

        options = yaml.load(open('./tests/files/test_options_ios.yml'))
        host = Host('example.rtr', 'ios', options)

        # Build HConfig object for the Running Config

        running_config_hier = HConfig(host=host)
        running_config_hier.load_from_file('./tests/files/running_config.conf')

        # Build Hierarchical Configuration object for the Compiled Config

        compiled_config_hier = HConfig(host=host)
        compiled_config_hier.load_from_file('./tests/files/compiled_config.conf')

        # Build Hierarchical Configuration object for the Remediation Config

        remediation_config_hier = running_config_hier.config_to_get_to(compiled_config_hier)

        for line in remediation_config_hier.all_children():
            print(line.cisco_style_text())

    See:

        ./tests/files/test_tags_ios.yml and ./tests/files/test_options_ios.yml

        for test examples of options and tags.

    """

    def __init__(self, hostname=None, os=None, options=None, host=None):
        if all([i is not None for i in (hostname, os, options)]):
            from hier_config.host import Host
            from warnings import warn
            self._host = Host(hostname, os, options)
            warning_message = """
            hostname, os, and options variables are being deprecated in version 2.0.0.
            Use the Host object going forward.
            Example:
                from hier_config.host import Host
                import yaml
                options = yaml.load(open('./tests/files/test_options_ios.yml'))
                host = Host({}, {}, options)
                hier = HConfig(host=host)""".format(hostname, os)
            warn(warning_message, SyntaxWarning)
        elif host is not None:
            assert hasattr(host, 'hostname')
            assert hasattr(host, 'os')
            assert hasattr(host, 'hconfig_options')
            self._host = host
        else:
            raise AttributeError('Error determining host object')

        self._options = self.host.hconfig_options
        self._text = str()
        self._logs = list()
        self.children = []
        self.children_dict = {}

    @property
    def host(self):
        return self._host

    @property
    def options(self):
        return self._options

    @property
    def logs(self):
        return self._logs

    @property
    def root(self):
        return self

    def __repr__(self):
        return 'HConfig(host={})'.format(self.host)

    def __str__(self):
        return self.text

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        if len(self.children) != len(other.children):
            return False

        for self_child, other_child in zip(sorted(self.children), sorted(other.children)):
            if self_child != other_child:
                return False
        return True

    def merge(self, other):
        """ Merges two HConfig objects """

        for child in other.children:
            self.add_deep_copy_of(child, merged=True)

    def load_from_file(self, file_path):
        """ Load configuration text from a file """

        with open(file_path) as f:
            config_text = f.read()
        self.load_from_string(config_text)

    def load_from_string(self, config_text):
        """ Create Hierarchical Configuration nested objects from text """

        for sub in self.options['full_text_sub']:
            config_text = re.sub(
                sub['search'],
                sub['replace'],
                config_text)

        current_section = self
        current_section.real_indent_level = -1
        most_recent_item = current_section
        indent_adjust = 0
        end_indent_adjust = []
        temp_banner = []
        in_banner = False
        banner_end_lines = ['EOF', '%', '!']
        banner_end_contains = []

        def end_of_banner_test(config_line):
            """
            :param config_line: type str
            :return: boolean
            """
            if config_line.startswith('^'):
                return True
            elif config_line in banner_end_lines:
                return True
            elif any([c in config_line for c in banner_end_contains]):
                return True
            return False

        for line in config_text.splitlines():
            # Process banners in configuration into one line
            if in_banner:
                if line != '!':
                    temp_banner.append(line)

                # Test if this line is the end of a banner
                if end_of_banner_test(str(line)):
                    in_banner = False
                    most_recent_item = self.add_child(
                        "\n".join(temp_banner), True)
                    most_recent_item.real_indent_level = 0
                    current_section = self
                    temp_banner = []
                continue
            else:
                # Test if this line is the start of a banner
                if line.startswith('banner '):
                    in_banner = True
                    temp_banner.append(line)
                    banner_words = line.split()
                    try:
                        banner_end_contains.append(banner_words[2])
                        banner_end_lines.append(banner_words[2][:1])
                        banner_end_lines.append(banner_words[2][:2])
                    except IndexError:
                        pass
                    continue

            actual_indent = len(line) - len(line.lstrip())
            line = ' ' * actual_indent + ' '.join(line.split())
            for sub in self.options['per_line_sub']:
                line = re.sub(
                    sub['search'],
                    sub['replace'],
                    line)
            line = line.rstrip()

            # If line is now empty, move to the next
            if not line:
                continue

            # Determine indentation level
            this_indent = len(line) - len(line.lstrip()) + indent_adjust

            line = line.lstrip()

            # Walks back up the tree
            while this_indent <= current_section.real_indent_level:
                current_section = current_section.parent

            # Walks down the tree by one step
            if this_indent > most_recent_item.real_indent_level:
                current_section = most_recent_item

            most_recent_item = current_section.add_child(line, True)
            most_recent_item.real_indent_level = this_indent

            for expression in self.options['indent_adjust']:
                if re.search(expression['start_expression'], line):
                    indent_adjust += 1
                    end_indent_adjust.append(expression['end_expression'])
                    break
            if end_indent_adjust and re.search(end_indent_adjust[0], line):
                indent_adjust -= 1
                del (end_indent_adjust[0])

        # Assert that we are not in a banner still for some reason
        assert not in_banner

        if self.host.os in ['ios']:
            self._remove_acl_remarks()
            self._add_acl_sequence_numbers()
            self._rm_ipv6_acl_sequence_numbers()

        return self

    def load_from_dump(self, dump):
        """
        Load a HConfig dump

        .. code:: python

            dump = [{
                'depth': child.depth(),
                'text': child.text,
                'tags': list(child.tags),
                'comments': list(child.comments),
                'new_in_config': child.new_in_config
            },]

        :param dump: list
        :return: None

        """
        from itertools import islice
        last_item = self
        for item in dump:
            # parent is the root
            if item['depth'] == 1:
                parent = self
            # has the same parent
            elif last_item.depth() == item['depth']:
                parent = last_item.parent
            # is a child object
            elif last_item.depth() + 1 == item['depth']:
                parent = last_item
            # has a parent somewhere closer to the root but not the root
            else:
                # last_item.lineage() = (a, b, c, d, e), new_item['depth'] = 2,
                # parent = a
                parent = next(islice(last_item.lineage(), item[
                              'depth'] - 2, item['depth'] - 1))
            # also accept 'line'
            # obj = parent.add_child(item.get('text', item['line']), force_duplicate=True)
            obj = parent.add_child(item['text'], force_duplicate=True)
            obj.tags = set(item['tags'])
            obj.comments = set(item['comments'])
            obj.new_in_config = item['new_in_config']
            last_item = obj

    def dump(self, lineage_rules=None):
        """
        Dump a list of loaded HConfig data

        .. code:: python

            dump = [{
                'depth': child.depth(),
                'text': child.text,
                'tags': list(child.tags),
                'comments': list(child.comments),
                'new_in_config': child.new_in_config
            },]

        :param lineage_rules: list
        :returns: list

        """

        if lineage_rules:
            children = self.all_children_sorted_with_lineage_rules(
                lineage_rules)
        else:
            children = self.all_children_sorted()

        output = []
        for child in children:
            output.append({
                'depth': child.depth(),
                'text': child.text,
                'tags': list(child.tags),
                'comments': list(child.comments),
                'new_in_config': child.new_in_config,
            })

        return output

    def add_tags(self, tag_rules, strip_negation=False):
        """
        Handler for tagging sections of Hierarchical Configuration data structure
        for inclusion and exclusion.

        """

        for rule in tag_rules:
            for child in self.all_children():
                if child.lineage_test(rule, strip_negation):
                    if 'add_tags' in rule:
                        child.deep_append_tags(rule['add_tags'])
                        for ancestor in child.lineage():
                            ancestor.append_tags(rule['add_tags'])
                    if 'remove_tags' in rule:
                        child.deep_remove_tags(rule['remove_tags'])

        return self

    def _idempotent_acl_check(self):
        """
        Handle conditional testing to determine if idempotent acl handling for iosxr should be used

        """

        return False

    def depth(self):
        return 0

    def _add_acl_sequence_numbers(self):
        """
        Add ACL sequence numbers for use on configurations with a style of 'ios'

        """

        ipv4_acl_sw = 'ip access-list'
        # ipv6_acl_sw = ('ipv6 access-list')
        if self.host.os in ['ios']:
            acl_line_sw = ('permit', 'deny')
        else:
            acl_line_sw = ('permit', 'deny', 'remark')
        for child in self.children:
            if child.text.startswith(ipv4_acl_sw):
                sn = 10
                for sub_child in child.children:
                    if sub_child.text.startswith(acl_line_sw):
                        sub_child.text = "{} {}".format(sn, sub_child.text)
                        sn += 10

        return self

    def _rm_ipv6_acl_sequence_numbers(self):
        """
        If there are sequence numbers in the IPv6 ACL, remove them

        """

        for acl in self.get_children('startswith', 'ipv6 access-list '):
            for entry in acl.children:
                if entry.text.startswith('sequence'):
                    entry.text = ' '.join(entry.text.split()[2:])
        return self

    def _remove_acl_remarks(self):
        for acl in self.get_children('startswith', 'ip access-list '):
            for entry in acl.children:
                if entry.text.startswith('remark'):
                    acl.children.remove(entry)
        return self

    def all_children_sorted_with_lineage_rules(self, rules):
        """

        Args:

            rules: list of lineage rules

        Yields:

            instances of Hierarchical Configuration that match one of the lineage rules

        """

        yielded = set()
        matched = set()
        for child in self.all_children_sorted():
            for ancestor in child.lineage():
                if ancestor in matched:
                    yield child
                    yielded.add(child)
                    break
            else:
                for rule in rules:
                    if child.lineage_test(rule, False):
                        matched.add(child)
                        for ancestor in child.lineage():
                            if ancestor in yielded:
                                continue
                            yield ancestor
                            yielded.add(ancestor)
                        break

    def add_ancestor_copy_of(self, parent_to_add):
        """
        Add a copy of the ancestry of parent_to_add to self
        and return the deepest child which is equivalent to parent_to_add

        :param parent_to_add: type HConfig
        :return: base

        """

        base = self
        for parent in parent_to_add.lineage():
            if parent.root is not parent:
                base = base.add_shallow_copy_of(parent)

        return base



import re


class TextMatch(object):
    """
    Provides a suite of text matching methods
    """

    @classmethod
    def dict_call(cls, test, text, expression):
        """
        Allows test methods to be called easily from variables
        """
        return {
            'equals': cls.equals,
            'startswith': cls.startswith,
            'endswith': cls.endswith,
            'contains': cls.contains,
            're_search': cls.re_search,
            'contains_or_endswith': cls.contains_or_endswith,
            'anything': cls.anything,
            'nothing': cls.nothing
        }[test](text, expression)

    @staticmethod
    def equals(text, expression):
        """Text equivalence test"""
        if isinstance(expression, str):
            return text == expression
        else:
            return text in expression

    @staticmethod
    def startswith(text, expression):
        """Text starts with test"""
        return text.startswith(expression)

    @staticmethod
    def endswith(text, expression):
        """Text ends with test"""
        return text.endswith(expression)

    @staticmethod
    def contains(text, expression):
        """Text contains test"""
        return expression in text

    @staticmethod
    def anything(text, expression):
        """Always returns True"""
        return True

    @staticmethod
    def nothing(text, expression):
        """Always returns False"""
        return False

    @staticmethod
    def contains_or_endswith(text, expression, with_pad=False):
        """Text contains test"""
        if with_pad:
            if TextMatch.endswith(text, ' {}'.format(expression)):
                return True
            elif TextMatch.contains(text, ' {} '.format(expression)):
                return True
        else:
            if TextMatch.endswith(text, expression):
                return True
            elif TextMatch.contains(text, expression):
                return True
        return False

    @staticmethod
    def re_search(text, expression):
        """
        Test regex match. This method is comparatively
        very slow and should be avoided where possible.
        """
        return re.search(expression, text) is not None



#!/usr/bin/env python

from setuptools import setup
from setuptools import find_packages
from hier_config import __version__

setup(
    name="hier_config",
    version=__version__,
    description="A router and switch configuration intention tool, used to build remediation configurations.",
    long_description="A router and switch configuration intention tool, used to build remediation configurations.",
    url="https://netdevops.io/hier_config/",
    license="MTI",
    packages=find_packages(exclude=['docs', 'tests']),
    author="Andrew Edwards, Jan Brooks, James Williams",
    author_email="andrew.edwards@rackspace.com, jan.brooks@rackspace.com, james.williams@rackspace.com",
    keywords="hier_config",
    python_requires='>=2.7',
    install_requires=[
        "PyYAML>=3.12",
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Telecommunications Industry',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Natural Language :: English',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Networking',
    ]
)
